library(refund)
library(mgcv)
library(lme4)
library(methods)
library(dplyr)
library(magrittr)
library(tidyr)
library(ggplot2)
library(rstan)
library(splines)

options(stringsAsFactors=F)

source("Utilities.R")
source("VB.R")
source("Stan.R")
source("Simulate.R")

# seed is which random seed to use, to generate different simulation replicates
# which.sim is which simulation scenario to use
# 1 is 24 subjects, 4 motions per subject
# 2 is 24 subjects, 12 motions per subject
# 3 is 24 subjects, 24 motions per subject (this simulation includes all three methods)
# 4 is 24 subjects, 48 motions per subject
# results will be saved in a folder called Simulations


RunSimulationByNumber <- function(seed, which.sim){
  s <- MakeSims()
  print(s[which.sim, ])
  num.subjects <- s$num.subjects[which.sim]
  num.motions.per.subject <- s$num.motions.per.subject[which.sim]
  group.var.type <- s$group.var.type[which.sim]
  prop.group.1 <- s$prop.group.1[which.sim]
  sim.Kp <- s$sim.Kp[which.sim]
  fit.Kp <- s$fit.Kp[which.sim]
  fit.Kt <- s$fit.Kt[which.sim]
  fix.sigma <- s$fix.sigma[which.sim]
  sim.error.sd <- s$sim.error.sd[which.sim]
  sim.name <- s$experiment[which.sim]
  STAN <- s$STAN[which.sim]
  PFFR <- s$PFFR[which.sim]
  RunSimulation(seed=seed, 
                num.subjects=num.subjects, 
                num.motions.per.subject=num.motions.per.subject,
                sim.Kp=sim.Kp, 
                fit.Kp=fit.Kp, fit.Kt=fit.Kt, 
                sim.error.sd=sim.error.sd, 
                sim.name=sim.name, 
                group.var.type=group.var.type, fix.sigma=fix.sigma, 
                prop.group.1=prop.group.1,
                STAN=STAN,
                PFFR=PFFR,
                D=50)
}


RunSimulation <- function(seed, 
                          num.subjects, 
                          num.motions.per.subject,
                          sim.Kp, 
                          fit.Kp, fit.Kt, 
                          sim.error.sd, max.iterations=100, 
                          sim.name, 
                          group.var.type, prop.group.1, fix.sigma, 
                          STAN, PFFR,
                          D=50, 
                          sim.alpha=0.0005)
{
  if (F){     # for testing
    num.subjects <- 8
    num.motions.per.subject <- 6
    D <- 20
    max.iterations <- 5
  }
  dir <- paste0("./Simulations/", sim.name, "/", seed, "/")
  dir.create(dir, showWarnings=F, recursive=T)
  sim.data.f <- paste0(dir, "simulated_data.Rdata")
  if (PFFR){
    mean_types <- c("pffr_ind", "pffr_dep", "vb")
  } else {
    mean_types <- "vb"
  }
  mean_types.df <- data.frame(Type=mean_types)
  mean_types.df$File <- paste0(dir, mean_types.df$Type, ".Rdata")
  combined.f <- paste0(dir, "vb_combined.Rdata")
  stan.f <- paste0(dir, "stan.Rdata")
  var_types <- c("glmer")
  var_types.df <- expand.grid(mean_types, var_types, stringsAsFactors=F)
  colnames(var_types.df) <- c("MeanType", "VarType")
  var_types.df$File <- paste0(dir, var_types.df$MeanType, "__", var_types.df$VarType, ".Rdata")
  if (!file.exists(sim.data.f)){
    re.spline.coef.var <- 0.0005
    pc.re.var <- c(3,1,0.3,0.1)
    d <- SimulateArtificialData(score.seed=seed, 
                                num.subjects=num.subjects,
                                num.motions.per.subject=num.motions.per.subject, 
                                prop.group.1=prop.group.1,
                                group.var.type=group.var.type,
                                error.sd=sim.error.sd, 
                                D=D, 
                                sim.Kp=sim.Kp,
                                pc.re.var=pc.re.var,
                                sim.alpha=sim.alpha,
                                re.spline.coef.var=re.spline.coef.var)
    Y <- d$observed
    Subject <- paste0("V", d$subject.indicator)
    Group <- d$group.indicator
    save(d, Y, Subject, Group, file=sim.data.f)
  }
  for (i in 1:nrow(mean_types.df)){
    if (nrow(mean_types.df)==0) next
    file_name <- mean_types.df$File[i]
    if (!file.exists(file_name)){
      load(sim.data.f) # Y, Subject, Group
      ptm <- proc.time()
      type <- mean_types.df$Type[i]
      cat("Mean estimation", type, "\n")
      type.env <- Mean.Fit(type=type, Y=Y, X=d$X, V=d$V, Subject=Subject, 
                           fit.Kt=fit.Kt, fit.Kp=fit.Kp, D=D, 
                  pffr.ind.file=mean_types.df$File[mean_types.df$Type=="pffr_ind"], 
                           max.iterations=max.iterations)
      time <- proc.time() - ptm
      with(type.env, 
           save(random.effects, fpcs, resid.before.fpcs, efunctions, 
           evalues, scores, alpha.estim, time,
           file=file_name)
           )
    }
  }
  for (i in 1:nrow(var_types.df)){
    if (nrow(var_types.df)==0) next
    file_name <- var_types.df$File[i]
    if (!file.exists(file_name)){
      ptm <- proc.time()
      mean_type <- var_types.df$MeanType[i]
      var_type <- var_types.df$VarType[i]
      cat("Variance estimation", mean_type, "__", var_type, "\n")
      mean_file <- mean_types.df$File[mean_types.df$Type==mean_type]
      load(mean_file)
      load(sim.data.f)
      if (var_type=="glmer"){
        var.fes <- vector('list', fit.Kp)
        var.res <- vector('list', fit.Kp)
        var.res.covs <- vector('list', fit.Kp)
        for (j in 1:ncol(scores)){
          df <- data.frame(RESPONSE=scores[, j]^2, GROUP=as.factor(Group), 
                           SUBJECT=as.factor(Subject))
          ggg <- glmer(RESPONSE ~ GROUP + (1 | SUBJECT), family=Gamma("log"), dat=df)
          over_dispersion <- overdisp_fun(ggg)[2]
          lme4.list <- ggg@optinfo$conv$lme4
          if (length(lme4.list)>0){
            warning.message <- ggg@optinfo$conv$lme4$messages
            conv.code <- ggg@optinfo$conv$lme4$code
            converged <- F
          } else {
            warning.message <- "None" 
            conv.code <- NA
            converged <- T
          }
          estimates <- cumsum(fixef(ggg))
          cov.matrix <- vcov(ggg) * 2 / over_dispersion
          sds <- c(sqrt(cov.matrix[1,1]), 
                   sqrt(sum(diag(cov.matrix)) + 2 * cov.matrix[1,2]))
          lower <- estimates - 1.96 * sds
          higher <- estimates + 1.96 * sds
          fpc <- paste0("FPC", j)
          var.fes[[j]] <- data.frame(FPC=fpc, Estimate=estimates, Group=c(1,2), SD=sds, 
                                     Lower2.5=lower, Higher97.5=higher, 
                                     Converged=converged, Warning=warning.message, 
                                     Code=conv.code,
                                     stringsAsFactors=F)
          res <- ranef(ggg)$SUBJECT[, 1]
          var.res[[j]] <- data.frame(FPC=fpc, Subject=rownames(ranef(ggg)$SUBJECT), Estimate=res)
          var.res.covs[[j]] <- data.frame(FPC=fpc, Var=VarCorr(ggg)[1]$SUBJECT[1,1])
        }
        var.fes <- do.call('rbind', var.fes)
        row.names(var.fes) <- NULL
        var.res <- do.call('rbind', var.res)
        row.names(var.res) <- NULL
        var.res.covs <- do.call('rbind', var.res.covs)
      }
      time <- proc.time() - ptm
      save(var.fes, var.res, var.res.covs, time=time, file=file_name)
      rm(var.fes, var.res, var.res.covs, time)
    }
  }
  if (!file.exists(combined.f)){
    ptm <- proc.time()
    load(sim.data.f)
    vb_mean_file <- mean_types.df$File[mean_types.df$Type=="vb"]
    vb_mean_envir <- my_load(vb_mean_file)
    alpha.estim <- vb_mean_envir$alpha.estim
    glmer_file <- var_types.df$File[var_types.df$MeanType=="vb" & var_types.df$VarType=="glmer"] 
    iw.l <- GetIWStuff(glmer_file, fix.sigma=fix.sigma, one.target=F)
    vb <- vb_mult_fpca_het(Y=Y, X=d$X, V=d$V, 
                              het.model.matrix=d$het.model.matrix,
                              FE.variance=iw.l$FE.variance,
                              max.iterations=max.iterations, 
                              alpha=alpha.estim,
                              Kt=fit.Kt, Kp=fit.Kp, 
                              iw.matrices=iw.l$iw.matrices, iw.dfs=iw.l$iw.dfs)
    vb.processed <- ProcessVB(vb, Y=Y)
    resid.before.fpcs <- vb.processed$resid.before.fpcs
    random.effects <- vb.processed$random.effects
    efunctions <- vb.processed$efunctions
    fpcs <- vb.processed$fpcs
    scores <- vb.processed$scores
    evalues <- vb.processed$evalues
    var.fes <- vb.processed$var.fes
    var.res <- vb.processed$var.res
    time <- proc.time() - ptm
    save(resid.before.fpcs, random.effects, efunctions, fpcs, scores, evalues, 
         var.fes, var.res, alpha.estim, time, file=combined.f)
  }
  if ((!file.exists(stan.f) & STAN)){
    ptm <- proc.time()
    load(sim.data.f)
    load(combined.f)
    dat <- list()
    dat$Phis <- t(efunctions)
    dat$YbeforeFPCs <- resid.before.fpcs
    dat$Y <- Y
    dat$D <- D
    dat$I <- length(unique(Subject))
    dat$N <- nrow(Y)
    dat$Kp <- fit.Kp
    dat$Kt <- fit.Kt
    dat$g_param <- 1
    diff.mat <- diff(diag(dat$D), diff=2)
    dat$Theta <- bs(1:D, df=dat$Kt, intercept=T)
    dat$Q <- t(dat$Theta) %*% t(diff.mat) %*% diff.mat %*% dat$Theta
    dat$P_mat <- 0.999 * dat$Q + 0.001 * diag(1, nrow=dat$Kt, ncol=dat$Kt)
    dat$V <- d$V
    dat$which_re <- as.numeric(gsub("V", "", Subject))
    g.df <- unique(data.frame(Subject=dat$which_re, Group=Group+1))
    dat$group <- g.df$Group
    glmer_file <- var_types.df$File[var_types.df$MeanType=="vb" & var_types.df$VarType=="glmer"] 
    iw.l <- GetIWStuff(glmer_file, fix.sigma=fix.sigma, one.target=F)
    dat$dfs <- unlist(iw.l$iw.dfs)
    dat$mats <- 1 / unlist(iw.l$iw.matrices)
    stan.res <- DoStan(dat, n.iter=800, n.warmup=400, n.chains=4, dir=dir)
    rhat <- stan.res$rhat
    var.fes <- stan.res$var.fes
    var.res <- stan.res$var.res
    time <- proc.time() - ptm
    save(rhat, var.fes, var.res, time, file=stan.f)
  }
}

GetPenMatrix <- function(SubjTar, correlation){
  subjtars <- unique(SubjTar)
  K <- length(subjtars)
  print(K)
  pen <<- matrix(0, nrow=K, ncol=K)
  rownames(pen) <- colnames(pen) <- subjtars
  subject <- unlist(strsplit(as.character(subjtars), "-", fixed=T))[c(T,F)]
  for (s in unique(subject)){
    w <- which(subject==s)
    for (i in w){
      for (j in w){
        pen[i, j] <- correlation
        if (i==j) pen[i, j] <- 1
      }
    }
  }
  penalty <<- solve(pen)
}

Mean.Fit <- function(type, Y, X=NULL, V=NULL, Subject=NULL, Number=NULL, 
                     SubjTar=NULL,
                     fit.Kt, fit.Kp=NULL, 
                     D=NULL, pffr.ind.file=NULL, 
                     max.iterations=100, tensor=F, 
                     correlation=NA, ig.a=1, ig.b=1){
  bas <- list(bs="bs", k=fit.Kt, m=c(3,2))
  dataM <- list(Y=Y, 
              Subject=as.factor(Subject))
  if (!is.na(correlation)){
    penalty <- GetPenMatrix(SubjTar=SubjTar, correlation)
    dataM$SubjTar <- as.factor(SubjTar)
    re.name <- "SubjTar"
  } else {
    re.name <- "Subject"
  }
  if (type=="pffr_ind"){
    if (tensor){
      dataM$Number <- Number
      if (tensor & !is.na(correlation)) stop("Tensor and correlation not 0 not supported")
      fit <- pffr(Y ~ c(0) + s(Number) + s(Subject, bs="re"), 
                  algorithm="bam", 
                  bs.int=bas, 
                  bs.yindex=bas, data=dataM)
    } else {
      if (is.na(correlation)){
        fit <- pffr(Y ~ c(0) + s(Subject, bs="re"), 
                    bs.int=bas, bs.yindex=bas,
              algorithm="bam", 
              data=dataM)
      } else {
        cat("k=208")
        fit <- pffr(Y ~ c(0) + s(SubjTar, bs="mrf", k=208, 
                         xt=list(penalty=penalty)),
                  bs.int=bas,
                  bs.yindex=bas, data=dataM, algorithm="bam")
      }
    }
    loglik <- logLik(fit)
    random.effects <- GetPffrRandomEffects(coef(fit, n2=D), "yindex.vec", 
                                           re.name=re.name)
    resid.before.fpcs <- residuals(fit)
    fitted.values <- fitted(fit)
    sc <- my.fpca.sc(resid.before.fpcs, center=F, 
                   npc=fit.Kp, nbasis=fit.Kt)
    efunctions <- sc$efunctions
    fpcs <- GetFPCs(efunctions)
    evalues <- sc$evalues
    scores <- sc$scores
    alpha.estim <- NA 
    l.alpha <- NA
  }
  if (type=="pffr_dep"){
    pffr.ind.env <- my_load(pffr.ind.file) 
    dataM$id <- factor(1:nrow(dataM$Y))
    # refit model with fpc-based residuals
    t <- 1:D
    samfrac <- 0.1
    ind.efunctions <- pffr.ind.env$efunctions  #[thin.seq, ]
    ind.evalues <- pffr.ind.env$evalues
    fit <- pffr(Y ~ c(0) + s(Subject, bs="re") + 
                  pcre(id=id, efunctions=ind.efunctions, 
                       evalues=ind.evalues, yind=t), 
                yind=t, bs.yindex=bas, 
                bs.int=bas, 
                data=dataM, 
                samfrac=samfrac,
                algorithm="bam")
    pffr.coef <- coef(fit, n1=D, n2=D)
    random.effects <- GetPffrRandomEffects(pffr.coef, "t.vec", re.name=re.name)
    # efunctions, evalues and fpcs comes from pffr_ind
    which.term <- which(names(pffr.coef$smterms)=="pcre(id,ind.efunctions,ind.evalues,t)")
    fpc.re.coef <- pffr.coef$smterms[[which.term]]$coef[, c("id.vec", "t.vec", "value")]
    pcef <- spread(fpc.re.coef, t.vec, value) %>% arrange(id.vec) %>% 
      dplyr::select(-id.vec) %>% as.matrix()
    fitted.values <- fitted(fit)
    resid.before.fpcs <- dataM$Y - (fitted.values - pcef)
    sc <- my.fpca.sc(resid.before.fpcs, center=F, 
                   npc=fit.Kp, nbasis=fit.Kt)
    efunctions <- sc$efunctions
    fpcs <- GetFPCs(sc$efunctions)
    evalues <- sc$evalues
    scores <- sc$scores
    alpha.estim <- NA
    l.alpha <- NA
    loglik <- logLik(fit)
  }
  if (type=="vb"){
    l.alpha <- EstimateAlpha(Y=Y, X=X, V=V, Kt=fit.Kt, Kp=fit.Kp, 
                             ig.a=ig.a, ig.b=ig.b)
    vb <- vb_mult_fpca_het(Y=Y, X=X, V=V, 
                         max.iterations=max.iterations, 
                         alpha=l.alpha$alpha.estim,
                         Kt=fit.Kt, Kp=fit.Kp, 
                         ig.alpha=ig.a, ig.beta=ig.b)
    l <- ProcessVB(vb, mean.only=T, Y=Y)
    resid.before.fpcs <- l$resid.before.fpcs
    fitted.values <- NA
    random.effects <- l$random.effects
    efunctions <- l$efunctions
    alpha.estim <- l.alpha$alpha.estim
    fpcs <- l$fpcs
    scores <- l$scores
    evalues <- l$evalues
    loglik <- NA
  }
  list(resid.before.fpcs=resid.before.fpcs, random.effects=random.effects, 
       efunctions=efunctions, 
       fpcs=fpcs, scores=scores, evalues=evalues, alpha.estim=alpha.estim,
       l.alpha=l.alpha,
       loglik=loglik, fitted.values=fitted.values)
}



overdisp_fun <- function(model) {
  ## number of variance parameters in 
  ##   an n-by-n variance-covariance matrix
  vpars <- function(m) {
    nrow(m)*(nrow(m)+1)/2
  }
  model.df <- sum(sapply(VarCorr(model),vpars))+length(fixef(model))
  rdf <- nrow(model.frame(model))-model.df
  rp <- residuals(model,type="pearson")
  Pearson.chisq <- sum(rp^2)
  prat <- Pearson.chisq/rdf
  pval <- pchisq(Pearson.chisq, df=rdf, lower.tail=FALSE)
  c(chisq=Pearson.chisq,ratio=prat,rdf=rdf,p=pval)
}

GetPffrRandomEffects <- function(pffr.coef, time.col.name, re.name){
  re.term <- paste0("s(", re.name, ")")
  which.term <- which(names(pffr.coef$smterms)==re.term)
  re <- pffr.coef$smterms[which.term][[1]]$coef %>%
    rename(Estimate=value) %>% dplyr::select(-se)
  colnames(re)[colnames(re)==time.col.name] <- "Time"
  re
}

GetFPCs <- function(efunctions){
  D <- nrow(efunctions)
  ret <- as.data.frame(efunctions) %>% 
    mutate(Time=1:D) %>% 
    gather(FPC, Estimate, -Time) %>%
    mutate(FPC=gsub("V", "FPC", FPC))
  ret
}

ProcessVB <- function(vb, variance.only=F, mean.only=F, Y=NULL){
  if (!mean.only){
    fit.Kp <- length(vb$g_mean_l)
    var.fes <- vector('list', fit.Kp)
    var.res <- vector('list', fit.Kp)
    for (j in 1:fit.Kp){
      estimates <- vb$g_mean_l[[j]][1:2]
      sds <- sqrt(diag(vb$g_cov_l[[j]][1:2, 1:2]))
      lower <- estimates - 1.96 * sds
      higher <- estimates + 1.96 * sds
      fpc <- paste0("FPC", j)
      var.fes[[j]] <- data.frame(FPC=fpc, Estimate=estimates, Group=c(1,2), SD=sds, 
                                 Lower2.5=lower, Higher97.5=higher, 
                                 Converged=T, Warning=NA, Code=NA,
                                 stringsAsFactors=F)
      
      res <- vb$g_mean_l[[j]][3:length(vb$g_mean_l[[j]])]
      var.res[[j]] <- data.frame(FPC=fpc, Subject=paste0("V", 1:length(res)), Estimate=res)
    }
    var.fes <- do.call('rbind', var.fes)
    row.names(var.fes) <- NULL
    var.res <- do.call('rbind', var.res)
    row.names(var.res) <- NULL
  }
  if (!variance.only){
    D <- ncol(Y)
    resid.before.fpcs <- Y - vb$fixef.cur - vb$ranef.cur
    resid <- Y - vb$fixef.cur - vb$ranef.cur - vb$pcaef.cur
    s2sum <- function(x){sum(x^2)}
    SSE.df <- data.frame(SSE.before.fpcs=apply(resid.before.fpcs, 1, s2sum), 
                         SSE.after.fpcs=apply(resid, 1, s2sum))
    SSE <- sum(resid^2)
    
    random.effects <- as.data.frame(vb$Theta %*% vb$mu.q.BZ) %>%
        mutate(Time=1:D) %>% gather(Subject, Estimate, -Time)
    
    efunctions <- vb$Theta %*% vb$mu.q.Bpsi
    fpcs <- GetFPCs(efunctions)
    scores <- vb$mu.q.C
    evalues <- apply(scores, 2, var)
  }
  ret <- list()
  if (!mean.only) {
    ret$var.fes=var.fes; ret$var.res=var.res
  }
  if (!variance.only){ 
    ret$resid.before.fpcs=resid.before.fpcs
    ret$random.effects=random.effects
    ret$efunctions=efunctions
    ret$fpcs=fpcs
    ret$scores=scores
    ret$evalues=evalues
    ret$SSE=SSE
    ret$SSE.df=SSE.df
  }
  return(ret)
}

GetIWStuff <- function(glmm_file, fix.sigma, one.target=T, nested=F){
  load(glmm_file)
  if (one.target){
    fit.Kp <- length(var.res.covs)
  } else {
    fit.Kp <- nrow(var.res.covs)
  }
  iw.matrices <- vector('list', fit.Kp)
  iw.dfs <- vector('list', fit.Kp)
  for (j in 1:fit.Kp){
    if (!one.target){
      re.cov <- matrix(var.res.covs[j, "Var"])
      num.effects <- 1
      multiplier <- 3
      re.name <- "RE1"
    } else {
      re.cov <- var.res.covs[[j]]
      if (nested){
        nested.re.cov <- nested.var.res.covs[[j]]
      }
      num.effects <- 2
      multiplier <- 5
      re.name <- "RE1_RE2"
    }
    iw.matrices[[j]] <- list(re.cov * multiplier)
    names(iw.matrices[[j]])[1] <- re.name
    if (nested){
      iw.matrices[[j]][[2]] <- nested.re.cov * multiplier
      names(iw.matrices[[j]])[2] <- paste0(re.name, "Nested")
    }
    iw.dfs[[j]] <- list(num.effects)
    names(iw.dfs[[j]])[1] <- re.name
    if (!one.target){
      if (re.cov < 1e-1){
        iw.matrices[[j]] <- list(matrix(num.effects))
        names(iw.matrices[[j]])[1] <- re.name
      }
    }
  }
  FE.variance <- data.frame(comp=c("FE1", "FE2"), val=fix.sigma)
  return(list(iw.matrices=iw.matrices, iw.dfs=iw.dfs, FE.variance=FE.variance))
}

my_load <- function(f){
  env <- environment()
  load(f, envir=env)
  env
}
