# X is the model matrix for the fixed effects in the mean model
#   1 column per fixed effect, 1 row per functional observations
# Y is the matrix of functional observations, 1 row per observation
# Kt is the number of spline basis functions
# Kp is the number of FPCs
# alpha is how much zeroth derivative penalty to use in the estimation
#   of random functional effects
# ig.alpha and ig.beta are hyperparameters for the variance parameters
#   for the fixed and random functional effects and FPCs
# het.model.matrix is the model matrix for the variance model
#   fixed effects should have column names like FE1_1, FE2_1
#   random effects should have column names like RE1_1, RE1_2, 
#   all random effects RE1 will come from the same distribution
# V is the model matrix for the random functional effects
# FE.variance are the variances for the fixed effects in the variance model
#   it should have columns comp and val, comp should be FE1, FE2, val=100, 100
# max.iterations is the maximum number of iterations
# tol is the tolerance, for the change in parameter estimates
# iw.matrices is list of lists, length Kp
#   each list is a list of matrices, with names RE1, with the inverse Wishart
#   matrix parameters
# iw.dfs is like iw.matrices, but has the degrees of freedom instead of the 
#   inverse Wishart matrix parameters
# 

vb_mult_fpca_het <- function (X=NULL, Y=NULL, Kt = 5, Kp = 2, 
          alpha = 1, ig.alpha=1, ig.beta=1, 
          het.model.matrix=NULL, 
          V=NULL, FE.variance=NULL,
          max.iterations=100, tol=5e-6, iw.matrices=NULL, iw.dfs=NULL,
          verbose=T, calculate.bic=F) 
{
  naive <- F
  scores_squared <- NULL
  bivariate <- F
  orthogonalize <- T
  cat("alpha=", alpha, "\n")
  estimate.pcs <- Kp > 0
  if (!estimate.pcs) Kp <- 1
  I = dim(X)[1]         # number of curves
  D = dim(Y)[2]         # length of curves
  p = dim(X)[2]         # number of predictors
  if (is.null(V)){      # if there are no random effects make dummy matrix
    ranefs <- F
    V <- matrix(0, nrow=I, ncol=1)
  } else {
    ranefs <- T
  }
  heteroskedastic <- !is.null(het.model.matrix)
  if (bivariate){
    dimTheta <- D / 2     # if bivariate, Theta is a patterned matrix
  } else {
    dimTheta <- D    
  }
  library(splines)
  Thetab = bs(1:dimTheta, df = Kt, intercept = TRUE, degree = 3)
  if (bivariate){
    Theta <- kronecker(diag(1, 2), Thetab)
    effKt <- Kt * 2
  } else {
    Theta <- Thetab
    effKt <- Kt
  }
  diff0 = diag(1, dimTheta)
  diff2 = matrix(rep(c(1, -2, 1, rep(0, dimTheta - 2)), 
                     dimTheta - 2)[1:((dimTheta - 2) * dimTheta)], 
                 dimTheta - 2, dimTheta, byrow = TRUE)
  P0 = t(Thetab) %*% t(diff0) %*% diff0 %*% Thetab
  #P0 <- diff0
  P2 = t(Thetab) %*% t(diff2) %*% diff2 %*% Thetab
  P.re.mat = alpha * P0 + (1 - alpha) * P2
  P.mat = P2
  Y.vec = as.vector(t(Y))
  obspts.vec = !is.na(Y.vec)
  Y.vec = Y.vec[obspts.vec]
  J = sum(obspts.vec)
  t.designmat.X = t(kronecker(X, Theta)[obspts.vec, ])
  XtX = matrix(0, effKt * p, effKt * p)
  sumXtX = matrix(0, effKt * p, effKt * p)
  for (i in 1:I) {
    obs.points = which(!is.na(Y[i, ]))
    X.cur = kronecker(matrix(X[i, ], nrow = 1, ncol = p), 
                      Theta)[obs.points, ]
    XtX = XtX + crossprod(X.cur)
    sumXtX = sumXtX + t(X.cur) %*% X.cur
  }
  b.q.lambda.BW = rep(ig.beta, p)
  mu.q.BZ <- matrix(0, nrow=effKt, ncol=ncol(V))
  colnames(mu.q.BZ) <- colnames(V)
  b.q.lambda.BZ <- ig.beta
  #sigma.q.Bpsi = vector("list", Kp)
  #for (k in 1:Kp) {
  #  sigma.q.Bpsi[[k]] = diag(1, effKt)
  #}
  sigma.q.Bpsi <- diag(1, nrow=effKt * Kp)
  mu.q.Bpsi = matrix(0, nrow = effKt, ncol = Kp)
  psi.cur = t(mu.q.Bpsi) %*% t(Theta)
  b.q.lambda.Bpsi = rep(ig.beta, Kp)
  sigma.q.C = vector("list", I)
  for (k in 1:I) {
    sigma.q.C[[k]] = diag(1, Kp)
  }
  set.seed(2)    # for reproducibility of results
  mu.q.C = matrix(rnorm(I * Kp, 0, 0.01), I, Kp)
  b.q.sigma.me = ig.beta
  if (bivariate){
    b.q.lambda.BZ <- rep(b.q.lambda.BZ, 2)
    b.q.lambda.Bpsi <- rep(b.q.lambda.Bpsi, 2)
    b.q.lambda.BW <- rep(b.q.lambda.BW, 2)
  }
  pcaef.cur = matrix(0, I, D)
  j <- 1 
  inv.sigma.me <- (ig.alpha + J /2)/(b.q.sigma.me)
  while (j < max.iterations) {
    pen.beta <- kronecker(diag((ig.alpha + Kt/2)/b.q.lambda.BW, 
                               length(b.q.lambda.BW)), P.mat)
    mean.cur = as.vector(t(pcaef.cur) + t(V %*% t(mu.q.BZ) %*% t(Theta)))[obspts.vec] 
    sigma.q.beta = solve(inv.sigma.me * 
                           XtX + 
                           pen.beta)
    mu.q.beta = matrix(sigma.q.beta %*% 
                         (inv.sigma.me * t.designmat.X %*% (Y.vec - mean.cur)), 
                       nrow = effKt, ncol = p)
    colnames(mu.q.beta) <- colnames(X)
    fixef.cur = as.matrix(X %*% t(mu.q.beta) %*% t(Theta))
    if (ranefs){
      exp.bT.Q.b <- 0
      if (bivariate) exp.bT.Q.b <- rep(exp.bT.Q.b, 2)
      pen.BZ <- kronecker(diag((ig.alpha + ncol(V) * 
                                  Kt/2)/b.q.lambda.BZ, 
                               length(b.q.lambda.BZ)), P.re.mat)
      for (i in 1:ncol(V)){
        which.curves <- which(V[, i]==1); num.curves <- length(which.curves)
        theta.mats <- vector('list', num.curves)
        for (curve in 1:num.curves){
          theta.mats[[curve]] <- Theta[!is.na(Y[which.curves[curve], ]), ]
        }
        design.mat <- do.call('rbind', theta.mats)
        sigma.q.BZ <- solve(inv.sigma.me * 
                              t(design.mat) %*% design.mat + 
                              pen.BZ)
        Y.subj <- Y[which.curves, ]
        obspts.vec <- !is.na(as.vector(t(Y.subj)))
        mean.cur = as.vector(t(fixef.cur[which.curves, ] + 
                               pcaef.cur[which.curves, ]))[obspts.vec]
        Y.subj.vec <- as.vector(t(Y.subj))[obspts.vec]
        mu.q.BZ[, i] = sigma.q.BZ %*% (inv.sigma.me * t(design.mat)) %*% 
                                            (Y.subj.vec-mean.cur)
        for (term in 1:2){
          indices <- (nrow(P.mat) * (term - 1) + 1):(nrow(P.mat) * term)
          exp.bT.Q.b[term] <- exp.bT.Q.b[term] + 
            t(mu.q.BZ[indices, i]) %*% P.re.mat %*% mu.q.BZ[indices, i] + 
            sum(diag(P.re.mat %*% sigma.q.BZ[indices, indices]))
          if (!bivariate) break
        }
      }
    }  
    ranef.cur = V %*% t(mu.q.BZ) %*% t(Theta)
    designmat = kronecker(mu.q.C, Theta)[obspts.vec, ]
    if (estimate.pcs){
      sigma.q.Bpsi = solve(kronecker(diag((ig.alpha + Kt/2)/b.q.lambda.Bpsi, 
                                          nrow=length(b.q.lambda.Bpsi)), 
                                     P.mat) + inv.sigma.me * 
                             f_sum(mu.q.c = mu.q.C, sig.q.c = sigma.q.C, theta = t(Theta), 
                                   obspts.mat = !is.na(Y)))
      mu.q.Bpsi = matrix(inv.sigma.me * sigma.q.Bpsi %*% 
                           f_sum2(y = Y, fixef = fixef.cur + ranef.cur, mu.q.c = mu.q.C, 
                                  kt = effKt, theta = t(Theta)), nrow = effKt, ncol = Kp)
      if (orthogonalize & Kp > 1){
        psi.cur <- svd(mu.q.C %*% t(mu.q.Bpsi) %*% t(Theta))$v[, 1:Kp]
        psi.cur <- t(psi.cur)
        if (j > 1){ 
          for (l in 1:Kp){
            sign.flip <- sign(sum(psi.cur[l, ] * last.psis[l, ]))
            psi.cur[l, ] <- psi.cur[l, ] * sign.flip
          }
        }
        last.psis = psi.cur
      } else {
        psi.cur = t(mu.q.Bpsi) %*% t(Theta)
      }
      library(dplyr)
      library(magrittr)
      library(tidyr)
      pcs.to.plot <- as.data.frame(t(psi.cur)) %>% set_colnames(paste0("FPC", 1:Kp)) %>%
        mutate(Time=1:D) %>% gather(FPC, Value, -Time)
      #library(ggplot2)
      #ggplot(pcs.to.plot, aes(x=Time, y=Value)) + geom_line() + facet_wrap(~FPC)
      #ggsave(paste0("PCsPlots/iteration", j, ".pdf"), height=3, width=12)
      if (j==1){ # & heteroskedastic){
        inverse.score.variances <- matrix(1, I, Kp)
      }
      for (subj in 1:I) {
        obs.points = which(!is.na(Y[subj, ]))
        Theta_i = t(Theta)[, obs.points]
        #if (is.null(het.model.matrix)){
        #  m <- diag(1, Kp, Kp)
        #} else {
        m <- diag(inverse.score.variances[subj, ], Kp, Kp)
        #}
        sigma.q.C[[subj]] = solve(m + 
                                    inv.sigma.me * 
                                    (f_trace(Theta_i = Theta_i, 
                                             Sig_q_Bpsi = sigma.q.Bpsi, 
                                             Kp = Kp, Kt = effKt) + 
              psi.cur[, obs.points, drop=F] %*% t(psi.cur[, obs.points, drop=F])))
        mu.q.C[subj, ] = inv.sigma.me * 
          sigma.q.C[[subj]] %*% as.matrix(psi.cur[, obs.points, drop=F]) %*% 
          (Y[subj, obs.points] - fixef.cur[subj, obs.points] - ranef.cur[subj, obs.points])
      }
      if (!heteroskedastic){
        g_mean_l <- g_cov_l <- iw_mean_l <- NULL
        for (k in 1:Kp){
          first.order.term <- mu.q.C[, k]^2
          second.order.term <- unlist(lapply(sigma.q.C, function(x){x[k,k]}))
          mu.C_squared <- first.order.term + second.order.term
          isv <- (1 + nrow(Y) / 2) / (1 + sum(mu.C_squared) / 2)
          #if (j > 5){
          inverse.score.variances[, k] <- isv
          #}
        }
      } else {
        var.p <- ncol(het.model.matrix)
        if (j==1) {
          prior_cov_l <- g_mean_l <- g_cov_l <- iw_mean_l <- p0 <- vector('list', Kp)
          for (k in 1:Kp){
            prior_cov_l[[k]] <- diag(1, var.p)
            p0[[k]] <- rep(0, var.p)
          }
        } else {
          p0 <- g_mean_l
        }
        for (k in 1:Kp){
          first.order.term <- mu.q.C[, k]^2
          second.order.term <- unlist(lapply(sigma.q.C, function(x){x[k,k]}))
          if (!naive){
            mu.C_squared <- first.order.term + second.order.term
          } else {
            mu.C_squared <- scores_squared[, k]
          }
          #print(k)
          gr <- GammaRegression(mu.C_squared=mu.C_squared, 
                                prior.cov=prior_cov_l[[k]],
                                p0=p0[[k]], 
                                het.model.matrix=het.model.matrix,
                                FE.variance=FE.variance, iw.matrices=iw.matrices[[k]], 
                                iw.dfs=iw.dfs[[k]])
          g_mean_l[[k]] <- gr$g_mean
          g_cov_l[[k]] <- gr$g_cov
          prior_cov_l[[k]] <- gr$het.model.prior.var
          iw_mean_l[[k]] <- gr$iw.mean
          inverse.score.variances[, k] <- gr$recip.exp.g.x
        }
      }
    }
    pcaef.cur = as.matrix(mu.q.C %*% psi.cur)
    resid = as.vector(Y - fixef.cur - pcaef.cur - ranef.cur)
    b.q.sigma.me = as.numeric(ig.beta + 
                                0.5 * (crossprod(resid[!is.na(resid)]) + 
                                         sum(diag(sumXtX %*% sigma.q.beta)) + 
                                         f_sum4(mu.q.c = mu.q.C, 
                                                sig.q.c = sigma.q.C, 
                                                mu.q.bpsi = mu.q.Bpsi, 
                                                sig.q.bpsi = sigma.q.Bpsi, 
                                                theta = Theta, 
                                                obspts.mat = !is.na(Y), 
                                                psi.cur=psi.cur)))
    inv.sigma.me <- (ig.alpha + J /2)/(b.q.sigma.me)
    for (term in 1:length(b.q.lambda.BW)) {
      mu.indices <- 1:(nrow(P.mat)) + 10 * (1 - term %% 2) * bivariate
      s.indices <- (nrow(P.mat) * (term - 1) + 1):(nrow(P.mat) * term)
      if (bivariate){
        col.num <- ceiling(term / 2)
      } else {
        col.num <- term
      }
      b.q.lambda.BW[term] = ig.beta + 0.5 * 
        (t(mu.q.beta[mu.indices, col.num]) %*% P.mat %*% 
          mu.q.beta[mu.indices, col.num] + 
           sum(diag(P.mat %*% sigma.q.beta[s.indices, s.indices])))
    }
    for (term in 1:length(b.q.lambda.Bpsi)) {
      mu.indices <- 1:nrow(P.mat) + 10 * (1 - term %% 2) * bivariate
      s.indices <- (nrow(P.mat) * (term - 1) + 1):(nrow(P.mat) * term)
      if (bivariate){
        col.num <- ceiling(term / 2)
      } else {
        col.num <- term
      }
      b.q.lambda.Bpsi[term] = ig.beta + 
        0.5 * (t(mu.q.Bpsi[mu.indices, col.num]) %*% P.mat %*% mu.q.Bpsi[mu.indices, col.num] + 
        sum(diag(P.mat %*% sigma.q.Bpsi[s.indices, s.indices])))
    }
    if (ranefs){
      for (term in 1:2){
        b.q.lambda.BZ[term] = as.numeric(ig.beta + 0.5 * exp.bT.Q.b[term])
        if (!bivariate) break
      }
    }
    current.objective.mean <- c(as.numeric(mu.q.beta, mu.q.Bpsi, mu.q.C))
    if (!estimate.pcs) {
      g_mean_l <- list(0)
      g_cov_l <- NULL
      iw_mean_l <- NULL
    }
    current.objective.var <- unlist(g_mean_l)
    #cat(".")
    if (j >= 2){
      diff.mean <- sum((current.objective.mean-old.objective.mean)^2)
      tol.mean <- diff.mean < tol
      if (!is.null(het.model.matrix)){
        diff.var <- sum((current.objective.var-old.objective.var)^2)
        tol.var <- diff.var < tol
      } 
      if (is.null(het.model.matrix)){
        if (tol.mean & j > 10) break
        if (verbose) cat("Diff.mean", diff.mean, "\n")
      } else {
        if (naive & tol.var) break
        if (tol.mean & tol.var & j > 10) break
        if (verbose) cat("Diff.mean", diff.mean, "Diff.var", diff.var, "\n")
      }
    }
    old.objective.mean <- current.objective.mean
    old.objective.var <- current.objective.var
    j <- j + 1
  }
  sigeps.pm = 1/inv.sigma.me
  
  if (calculate.bic){
    n.curves <- I
    if (nrow(V) > 576){
      n.sub <- min(12, ncol(V))
      which.curves <- which(rowSums(V[, 1:n.sub])>0)
      n.curves <- length(which.curves)
    } else {
      which.curves <- 1:nrow(V)
      n.sub <- ncol(V)
    }
    fitted <- fixef.cur[which.curves, ] + pcaef.cur[which.curves, ] + ranef.cur[which.curves, ]
    omega.intercept <- X[which.curves, ] %x% Theta
    omega.res <- V[which.curves, 1:n.sub] %x% Theta
    omega.pcs <- diag(n.curves) %x% t(psi.cur)
    omega <- cbind(omega.intercept, omega.res, omega.pcs)
    vecY <- as.numeric(Y[which.curves, ])
    Pmat <- bdiag(pen.beta, diag(ncol(V[which.curves, 1:n.sub])) %x% pen.BZ, diag(as.numeric(t(inverse.score.variances[which.curves, ]))))
    omega.sparse <- Matrix(omega, sparse = TRUE)
    Pmat.sparse <- Matrix(Pmat, sparse = TRUE)
    A.sparse <- solve(crossprod(omega.sparse) + Pmat.sparse, t(omega.sparse))
    trace.hat <- sum(A.sparse * t(omega.sparse))
    log.sum.resid <- log(sum((fitted - Y[which.curves, ])^2))
    BIC <- n.curves * D * log.sum.resid + trace.hat * log(n.curves * D)
  } else {
    BIC <- NULL
  }
  
  ret = list(sigeps.pm=sigeps.pm, 
             psi.cur=psi.cur, 
             g_mean_l=g_mean_l, g_cov_l=g_cov_l, iw_mean_l=iw_mean_l, 
             mu.q.beta=mu.q.beta, mu.q.Bpsi=mu.q.Bpsi, mu.q.BZ=mu.q.BZ,
            sigma.q.beta=sigma.q.beta, sigma.q.Bpsi=sigma.q.Bpsi, 
            mu.q.C=mu.q.C, Theta=Theta, het.model.matrix=het.model.matrix,
            Kp=Kp, Kt=effKt, bivariate=bivariate, 
            heteroskedastic=heteroskedastic, 
            alpha=alpha,
            Y=Y, ranef.cur=ranef.cur, fixef.cur=fixef.cur, pcaef.cur=pcaef.cur, 
            X=X, V=V, BIC=BIC)
  ret
}

f_sum = function(mu.q.c, sig.q.c, theta, obspts.mat){
  I = dim(mu.q.c)[1]
  kp = dim(mu.q.c)[2]
  kt = dim(theta)[1]
  ret.sum = matrix(0, kp*kt, kp*kt)
  for(i in 1:I){
    mu.mat = matrix(mu.q.c[i,], nrow = 1, ncol = kp)
    ret.sum = ret.sum + kronecker(t(mu.mat) %*% mu.mat + sig.q.c[[i]], (theta[,obspts.mat[i,]])%*%t(theta[,obspts.mat[i,]])) 
  }
  return(ret.sum)
}

f_sum2 = function(y, fixef, mu.q.c, kt, theta){
  I = dim(mu.q.c)[1]
  kp = dim(mu.q.c)[2]
  ret.sum = matrix(0, nrow = kp*kt, ncol = 1)
  for(i in 1:I){
    obs.pts = !is.na(y[i,])
    ret.sum = ret.sum + kronecker((matrix(mu.q.c[i,])), theta[,obs.pts]) %*% matrix(y[i, obs.pts] - fixef[i,obs.pts])
  }
  return(ret.sum)
}

f_sum4 = function(mu.q.c, sig.q.c, mu.q.bpsi, sig.q.bpsi, theta, obspts.mat, psi.cur){
  I = dim(mu.q.c)[1]
  kp = dim(mu.q.c)[2]
  kt = dim(theta)[2]
  ret.sum = matrix(0, 1, 1)
  for(i in 1:I){
    theta_i = t(theta)[,obspts.mat[i,]]
    t.mu.q.bpsi_theta_i <- psi.cur[, obspts.mat[i, ], drop=F]
    temp = 
      f_trace(Theta_i = theta_i, Sig_q_Bpsi = sig.q.bpsi, Kp = kp, Kt = kt) %*% matrix(mu.q.c[i,], kp, 1) %*% matrix(mu.q.c[i,], 1, kp) +
      f_trace(Theta_i = theta_i, Sig_q_Bpsi = sig.q.bpsi, Kp = kp, Kt = kt) %*% sig.q.c[[i]] +
      t.mu.q.bpsi_theta_i %*% t(t.mu.q.bpsi_theta_i) %*% sig.q.c[[i]]
    ret.sum = ret.sum + sum(diag(temp))   
  }
  return(ret.sum)
}

f_trace = function(Theta_i, Sig_q_Bpsi, Kp, Kt){
  ret.mat = matrix(NA, nrow = Kp, ncol = Kp)
  A = Theta_i %*% t(Theta_i)
  for(i in 1:Kp){
    for(j in 1:Kp){
      ret.mat[i,j] = sum(diag(A %*% Sig_q_Bpsi[((-1 + i)*Kt + 1):(i*Kt), ((-1 + j)*Kt + 1):(j*Kt)]))
    }
  }
  return(ret.mat)
}

GammaRegression <- function(mu.C_squared, prior.cov=100, p0, 
                            het.model.matrix,
                            FE.variance=NULL, iw.matrices=NULL, iw.dfs=NULL){
  #iw.nu <- 2
  #iw.psi <- diag(iw, nrow=2)
  source("vbhetlm1.R")
  mu.C_squared <- matrix(mu.C_squared)
  prior.mean <- matrix(rep(0, ncol(het.model.matrix)))
  fit <- bgamglm(mu.C_squared, het.model.matrix, 
                 prior.mean, prior.cov, 
                 p0=p0, tol=1e-6)
  fit.recip.exp.g.x <- exp(
    -het.model.matrix %*% fit$coef 
    -1/2 * diag(het.model.matrix %*% 
                  fit$variance %*% 
                  t(het.model.matrix) ))
  g_mean <- as.numeric(fit$coef)
  recip.exp.g.x <- as.numeric(fit.recip.exp.g.x)
  g_cov <- fit$variance
  
  # colnames of het.model.matrix must be of form FE1_1, FE2_1, RE1_1, RE1_2
  # set up het.model.prior.var
  # only works if all random effects have a single prior distribution
  cnhmm <- colnames(het.model.matrix)
  names(g_mean) <- colnames(g_cov) <- rownames(g_cov) <- cnhmm
  het.model.prior.var <- 
    matrix(0, nrow=ncol(het.model.matrix), 
           ncol=ncol(het.model.matrix), 
           dimnames=list(cnhmm, cnhmm))
  colnames(het.model.prior.var) <- row.names(het.model.prior.var) <- colnames(het.model.matrix)
  effects <- unlist(strsplit(cnhmm, "_"))[c(T,F)]
  effects.type <- ifelse(grepl("FE", effects), "fixed", "random")
  effects.subject <- unlist(strsplit(cnhmm, "_"))[c(F,T)]
  effects.subject[effects.type=="fixed"] <- NA
  u.effects <- unique(effects)
  # first do fixed effects
  fixed.effects <- u.effects[grepl("FE", u.effects)]
  random.effects <- u.effects[grepl("RE", u.effects)]
  for (fe in fixed.effects){
    w <- which(effects==fe)
    het.model.prior.var[cbind(w, w)] <- FE.variance$val[FE.variance$comp==fe]
  }
  if (length(random.effects)>0){
    subjects <- effects.subject[effects.type=="random"]
    n.subjects <- length(unique(subjects))
    n.res <- length(random.effects)
    matrix.sum <- matrix(0, n.res, n.res)
    for (subj in unique(subjects)){
      w.subj <- which(effects.subject==subj & effects.type=="random")
      subj.mean <- g_mean[w.subj]
      subj.cov <- g_cov[w.subj, w.subj]
      matrix.sum <- matrix.sum + matrix(subj.mean, ncol=1) %*% matrix(subj.mean, nrow=1) + subj.cov
    }
    p <- nrow(iw.matrices[[1]])
    iw.mean <- (iw.matrices[[1]] + matrix.sum) / (iw.dfs[[1]] + n.subjects - p - 1)
    for (subj in subjects){
      w.subj <- which(effects.subject==subj & effects.type=="random")
      het.model.prior.var[w.subj, w.subj] <- as.matrix(iw.mean)
    }
  } else {
    iw.mean <- NULL
  }
  return(list(g_mean=g_mean, g_cov=g_cov, recip.exp.g.x=recip.exp.g.x, 
              het.model.prior.var=het.model.prior.var, 
              iw.mean=iw.mean))
}

EstimateAlpha <- function(Y, X, V, Kt, Kp, bivariate=F, ig.a=1, ig.b=1){
  alpha.1 <- 10^-(2:8)
  alpha.2 <- 1:10/10
  alpha.grid <- sort(c(alpha.1, alpha.2))
  num.alphas <- length(alpha.grid)
  res <- data.frame(alpha=alpha.grid, BIC=NA, exclude=F)
  to.try <- sample(1:num.alphas, num.alphas)
  min.to.try <- 0
  max.to.try <- 1
  for (i in 1:length(to.try)){
    if (res$exclude[to.try[i]]) next
    res$BIC[to.try[i]] <- FitRandomEffects(res$alpha[to.try[i]], 
                                           Y=Y, X=X, V=V, Kt=Kt, Kp=Kp, 
                                           ig.a=ig.a, ig.b=ig.b)
    if (i >= 2){
      which.min <- which.min(res$BIC)
      alpha.min <- res$alpha[which.min]
      others <- which(res$BIC>which.min)
      for (j in others){
        alpha.other <- res$alpha[j]
        if (alpha.other < alpha.min){
          res$exclude[res$alpha <= alpha.other] <- T
        } else if (alpha.other > alpha.min) {
          res$exclude[res$alpha >= alpha.other] <- T
        }
      }
    }
  }
  alpha.estim <- res$alpha[which.min(res$BIC)]
  return(list(res=res, alpha.estim=alpha.estim))
}


FitRandomEffects <- function(est.alpha, Y, X, V, Kt, Kp, bivariate=F, ig.a=1, ig.b=1){
  ff <- 
    vb_mult_fpca_het(Y=Y, X=X, V=V, 
                     Kt=Kt, Kp=Kp,
                     alpha=est.alpha, verbose=F, max.iterations=11, 
                     ig.alpha=ig.a, ig.beta=ig.b, calculate.bic=T)
  return(ff$BIC)
}

